using System;
using System.Collections.Generic;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Security;
using System.Threading;
using System.Windows.Forms;

namespace MyToolTipLib {
    [SuppressUnmanagedCodeSecurityAttribute]
    internal static class SafeNativeMethods {
        internal static IntPtr MAKELONG(short wLow, short wHigh) { return (IntPtr)((wLow & 0xffff) | (wHigh & 0xffff) << 16); }

        [Flags] //WinUser.h
        internal enum WindowStyleFlags : uint {
            WM_USER = 0x0400,
            WS_EX_TOPMOST = 0x0008,
            CW_USEDEFAULT = 0x80000000
        }
        [Flags] //CommCtrl.h
        internal enum ToolTipControlFlags : uint {
            TTS_ALWAYSTIP = 0x01,
            TTS_NOPREFIX = 0x02,
            TTF_TRACK = 0x0020,
            TTF_ABSOLUTE = 0x0080,
            TTM_ADDTOOLA = ((ushort)WindowStyleFlags.WM_USER + 4),
            TTM_DELTOOLA = ((ushort)WindowStyleFlags.WM_USER + 5),
            TTM_UPDATETIPTEXTA = ((ushort)WindowStyleFlags.WM_USER + 12),
            TTM_TRACKACTIVATE = ((ushort)WindowStyleFlags.WM_USER + 17),  // wParam = TRUE/FALSE start end  lparam = LPTOOLINFO
            TTM_TRACKPOSITION = ((ushort)WindowStyleFlags.WM_USER + 18),  // lParam = dwPos
            TTM_SETMAXTIPWIDTH = ((ushort)WindowStyleFlags.WM_USER + 24),
            TTM_GETBUBBLESIZE = ((ushort)WindowStyleFlags.WM_USER + 30)
        }

        /*LRESULT SendMessage(
             HWND hWnd,
             UINT Msg,
             WPARAM wParam,
             LPARAM lParam
             );*/
        [DllImport("user32.dll")]
        internal static extern IntPtr SendMessage(IntPtr hWnd, ToolTipControlFlags msg, IntPtr wParam, IntPtr lParam);

        /*LRESULT SendMessage(
             HWND hWnd,
             UINT Msg,
             WPARAM wParam,
             TOOLINFO lParam
             );*/
        [DllImport("user32.dll")]
        internal static extern IntPtr SendMessage(IntPtr hWnd, ToolTipControlFlags msg, IntPtr wParam, ref TOOLINFO lParam);

        /*BOOL ScreenToClient(
            HWND hWnd,
            LPPOINT lpPoint
        );*/
        [DllImport("user32.dll")]
        [return: MarshalAs(UnmanagedType.Bool)]
        internal static extern bool ClientToScreen(IntPtr hWnd, ref Point lpPoint);

        [StructLayout(LayoutKind.Sequential)]
        internal struct TOOLINFO {
            public uint cbSize;
            public ToolTipControlFlags uFlags;
            public IntPtr hWnd;
            public IntPtr uId;
            public RECT rect;
            public IntPtr hinst;
            public string strText;
            public IntPtr lParam;
        }

        [StructLayout(LayoutKind.Sequential)]
        internal struct RECT {
            public int left;
            public int top;
            public int right;
            public int bottom;
        }
    }

    public static class MyToolTipManager {
        private class myReverserSortClass : IComparer<ushort> { int IComparer<ushort>.Compare(ushort x, ushort y) { return x == y ? 0 : (x > y ? -1 : 1); } }
        private static readonly object myToolTipLock = new object();
        private static Dictionary<IntPtr, ushort> valCheck = new Dictionary<IntPtr, ushort>(5);  //tipWin / Start
        private static SortedDictionary<ushort, ushort> valMain;  //Start / End
        private static SortedDictionary<ushort, ushort> valMainReverse;  //Start / End

        public static ushort GetNextSlotLeft(IntPtr tipWin, ushort width) {
            lock(myToolTipLock) {
                if(valMain == null) valMain = new SortedDictionary<ushort, ushort>();
                else if(valCheck.ContainsKey(tipWin)) return valCheck[tipWin];

                ushort start = 0;
                foreach(KeyValuePair<ushort, ushort> A in valMain) {
                    if(A.Key - start >= width) break;
                    else start = A.Value;
                }

                valCheck[tipWin] = start;
                valMain[start] = (ushort)(start + width);
                return start;
            }
        }
        public static ushort GetNextSlotRight(IntPtr tipWin, ushort width) {
            lock(myToolTipLock) {
                if(valMainReverse == null) valMainReverse = new SortedDictionary<ushort, ushort>(new myReverserSortClass());
                if(valCheck.ContainsKey(tipWin)) return valCheck[tipWin];

                ushort end = 620;
                foreach(KeyValuePair<ushort, ushort> A in valMainReverse) {
                    if(end - A.Value >= width) break;
                    else end = A.Key;
                }

                valCheck[tipWin] = (ushort)(end - width);
                valMainReverse[(ushort)(end - width)] = end;
                return (ushort)(end - width);
            }
        }
        public static void RemoveSlot(IntPtr tipWin) {
            lock(myToolTipLock)
                if(valCheck.ContainsKey(tipWin)) {
                    if(valMain != null) valMain.Remove(valCheck[tipWin]);
                    if(valMainReverse != null) valMainReverse.Remove(valCheck[tipWin]);
                    valCheck.Remove(tipWin);
                }
        }
        public static bool HasTip(IntPtr tipWin) { lock(myToolTipLock) return valCheck.ContainsKey(tipWin); }
    }

    public class MyToolTip : NativeWindow, IDisposable {
        private Point toolTipPos = new Point(0, 0);
        private SafeNativeMethods.TOOLINFO ti = new SafeNativeMethods.TOOLINFO();

        public MyToolTip(string text, int displayTimeMillis) { ShowToolTip(text, -1, -1, displayTimeMillis, IntPtr.Zero); }
        public MyToolTip(string text, int displayTimeMillis, IntPtr tipWin) { ShowToolTip(text, -1, -1, displayTimeMillis, tipWin); }
        public MyToolTip(string text, int locX, int locY, int displayTimeMillis) { ShowToolTip(text, locX, locY, displayTimeMillis, IntPtr.Zero); }
        public MyToolTip(string text, int locX, int locY, int displayTimeMillis, IntPtr tipWin) { ShowToolTip(text, locX, locY, displayTimeMillis, tipWin); }

        public MyToolTip(string text) : this() { ShowToolTip(text, -1, -1, IntPtr.Zero); }
        public MyToolTip(string text, IntPtr tipWin) : this() { ShowToolTip(text, -1, -1, tipWin); }
        public MyToolTip(string text, int locX, int locY) : this() { ShowToolTip(text, locX, locY, IntPtr.Zero); }
        public MyToolTip(string text, int locX, int locY, IntPtr tipWin) : this() { ShowToolTip(text, locX, locY, tipWin); }
        public MyToolTip() {
            ti.cbSize = (uint)Marshal.SizeOf(typeof(SafeNativeMethods.TOOLINFO));
            ti.uFlags = SafeNativeMethods.ToolTipControlFlags.TTF_TRACK | SafeNativeMethods.ToolTipControlFlags.TTF_ABSOLUTE;
            ti.hWnd = IntPtr.Zero;
            ti.uId = IntPtr.Zero;
            ti.rect.left = ti.rect.top = ti.rect.right = ti.rect.bottom = 0;
            ti.hinst = IntPtr.Zero;
            ti.strText = "";

            CreateParams cp = new CreateParams();
            cp.ExStyle = (int)SafeNativeMethods.WindowStyleFlags.WS_EX_TOPMOST;
            cp.ClassName = "tooltips_class32";
            cp.Style = (int)(SafeNativeMethods.ToolTipControlFlags.TTS_NOPREFIX | SafeNativeMethods.ToolTipControlFlags.TTS_ALWAYSTIP);
            cp.X = unchecked((int)SafeNativeMethods.WindowStyleFlags.CW_USEDEFAULT);
            cp.Y = unchecked((int)SafeNativeMethods.WindowStyleFlags.CW_USEDEFAULT);
            cp.Width = unchecked((int)SafeNativeMethods.WindowStyleFlags.CW_USEDEFAULT);
            cp.Height = unchecked((int)SafeNativeMethods.WindowStyleFlags.CW_USEDEFAULT);
            cp.Parent = IntPtr.Zero;
            cp.Param = IntPtr.Zero;
            base.CreateHandle(cp);

            SafeNativeMethods.SendMessage(this.Handle, SafeNativeMethods.ToolTipControlFlags.TTM_ADDTOOLA, IntPtr.Zero, ref ti);
            SafeNativeMethods.SendMessage(this.Handle, SafeNativeMethods.ToolTipControlFlags.TTM_SETMAXTIPWIDTH, IntPtr.Zero, (IntPtr)Screen.PrimaryScreen.Bounds.Right);
        }

        //If time is given, it'll automatically make the tooltip dissapear, else have to do manually with "".  Only the thread who made the "CreateHandle(cp)" window seems to be able to send to it... which is why the "Thread" makes a new tooltip.
        //Maybe better to not make window in the constructor and just keep making the window at the "showToolTip"...  To make the "ShowToolTip" with null string be able to make one dissapear would need a thread safe array and check if anything (with locX,locY) already exsists.  Would be able to Destory the window when the null string one is called.
        public void ShowToolTip(string text, int locX, int locY, int displayTimeMillis) { ShowToolTip(text, locX, locY, displayTimeMillis, IntPtr.Zero); }
        public void ShowToolTip(string text, int locX, int locY, int displayTimeMillis, IntPtr tipWin) {
            Thread startTool = new Thread(new ParameterizedThreadStart(ShowToolTip));
            startTool.Start(new object[] { text, locX, locY, displayTimeMillis, tipWin });
        }
        private void ShowToolTip(object myObj) {
            object[] myObjArr = myObj as object[];
            using(new MyToolTip((string)myObjArr[0], (int)myObjArr[1], (int)myObjArr[2], (IntPtr)myObjArr[4])) Thread.Sleep((int)myObjArr[3]);
        }

        public void ShowToolTip(string text) { ShowToolTip(text, -1, -1, IntPtr.Zero); }
        public void ShowToolTip(string text, IntPtr tipWin) { ShowToolTip(text, -1, -1, tipWin); }
        public void ShowToolTip(string text, int locX, int locY) { ShowToolTip(text, locX, locY, IntPtr.Zero); }
        public void ShowToolTip(string text, int locX, int locY, IntPtr tipWin) {
            if(locX >= 0 && locY >= 0) { toolTipPos.X = locX; toolTipPos.Y = locY; } else if(MyToolTipManager.HasTip(this.Handle)) { toolTipPos.X = MyToolTipManager.GetNextSlotRight(this.Handle, 0); toolTipPos.Y = 0; locX = locY = -2; }
            if(locX == -1 || locY == -1) { toolTipPos.X = 0; toolTipPos.Y = -18; } else if(tipWin != IntPtr.Zero) SafeNativeMethods.ClientToScreen(tipWin, ref toolTipPos);

            if(toolTipPos.X >= 0 && toolTipPos.Y >= -18) {
                ti.strText = text;
                SafeNativeMethods.SendMessage(this.Handle, SafeNativeMethods.ToolTipControlFlags.TTM_UPDATETIPTEXTA, IntPtr.Zero, ref ti);
                SafeNativeMethods.SendMessage(this.Handle, SafeNativeMethods.ToolTipControlFlags.TTM_TRACKPOSITION, IntPtr.Zero, SafeNativeMethods.MAKELONG((short)toolTipPos.X, (short)toolTipPos.Y));
                SafeNativeMethods.SendMessage(this.Handle, SafeNativeMethods.ToolTipControlFlags.TTM_TRACKACTIVATE, new IntPtr(1), ref ti);
            }

            if(locX == -2 && locY == -2 && string.IsNullOrEmpty(text)) MyToolTipManager.RemoveSlot(this.Handle);
            else if(locX == -1 && locY == -1 && !MyToolTipManager.HasTip(this.Handle)) {
                toolTipPos.X = MyToolTipManager.GetNextSlotRight(this.Handle, (ushort)((uint)SafeNativeMethods.SendMessage(this.Handle, SafeNativeMethods.ToolTipControlFlags.TTM_GETBUBBLESIZE, IntPtr.Zero, ref ti) & 0xffff)); toolTipPos.Y = 0;
                if(tipWin != IntPtr.Zero) SafeNativeMethods.ClientToScreen(tipWin, ref toolTipPos);

                if(toolTipPos.X >= 0 && toolTipPos.Y >= 0) {
                    SafeNativeMethods.SendMessage(this.Handle, SafeNativeMethods.ToolTipControlFlags.TTM_TRACKPOSITION, IntPtr.Zero, SafeNativeMethods.MAKELONG((short)toolTipPos.X, (short)toolTipPos.Y));
                    SafeNativeMethods.SendMessage(this.Handle, SafeNativeMethods.ToolTipControlFlags.TTM_TRACKACTIVATE, new IntPtr(1), ref ti);
                }
            }
        }

        //Note: Should add Finalizer so tooltip cleans up even if user forgot, but need to be able to call ShowToolTip only when its needed (Mainly because of the timed tooltips)
        private bool disposed;
        public void Dispose() { Dispose(true); GC.SuppressFinalize(this); }
        protected virtual void Dispose(bool disposing) {
            if(!disposed) {
                if(disposing) {
                    ShowToolTip("");
                    if(base.Handle != null) base.DestroyHandle();
                }
                disposed = true;
            }
        }
    }
}